from dataclasses import dataclass


@dataclass(frozen=True)
class WorfError(Exception):
    message: str


@dataclass(frozen=True)
class ActionError(WorfError):
    message: str


@dataclass(frozen=True)
class AuthenticationError(WorfError):
    message: str = "Unauthorized"


@dataclass(frozen=True)
class DataConflict(WorfError):
    message: str = "Conflict"


@dataclass(frozen=True)
class NamingThingsError(WorfError, ValueError):
    message: str


@dataclass(frozen=True)
class NotFound(WorfError):
    message: str = "Not found"


@dataclass(frozen=True)
class PermissionsError(WorfError):
    message: str


@dataclass(frozen=True)
class FieldError(WorfError, ValueError):
    message: str



from datetime import timedelta
from unittest.mock import patch
from uuid import uuid4

from django.core.files.uploadedfile import SimpleUploadedFile

from tests import parametrize


def test_profile_detail(client, db, profile, user):
    response = client.get(f"/profiles/{profile.pk}/")
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == user.username


def test_profile_not_found(client, db, profile, user):
    response = client.get(f"/profiles/{uuid4()}/")
    result = response.json()
    assert response.status_code == 404, result
    assert result["message"] == "Not found"


def test_profile_delete(client, db, profile, user):
    response = client.delete(f"/profiles/{profile.pk}/")
    assert response.status_code == 204, response.content
    assert response.content == b""


def test_profile_list(client, db, profile, user):
    response = client.get("/profiles/")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    assert result["profiles"][0]["username"] == user.username


@parametrize("page", [-1, 0, 1, 2])
def test_profile_list_pages(client, db, page):
    response = client.get("/profiles/", dict(page=page))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 0


def test_profile_list_filters(client, db, profile, url, user):
    response = client.get(url("/profiles/", {"name": user.name}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    assert result["profiles"][0]["username"] == user.username


@parametrize("url_params__array_format", ["repeat"])  # comma fails on ands
def test_profile_list_and_filters(client, db, profile_factory, tag_factory, url):
    tag1, tag2, tag3 = tag_factory.create_batch(3)
    profile_factory.create(tags=[tag1])
    profile_factory.create(tags=[tag2])
    profile_factory.create(tags=[tag1, tag2])
    profile_factory.create(tags=[tag3])
    profile_factory.create()
    response = client.get(url("/profiles/", {"tags": [tag1.pk, tag2.pk]}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1


def test_profile_list_annotation_filters(client, db, profile_factory, url):
    profile_factory.create(user__date_joined="2020-01-01T00:00:00Z")
    profile_factory.create(user__date_joined="2020-12-01T00:00:00Z")
    filters = {"dateJoined__gte": "2020-06-01T00:00:00Z"}
    response = client.get(url("/profiles/", filters))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1


def test_profile_list_icontains_filters(client, db, profile, url, user):
    response = client.get(url("/profiles/", {"name__icontains": user.first_name}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    assert result["profiles"][0]["username"] == user.username


@parametrize("url_params__array_format", ["comma", "repeat"])
def test_profile_list_in_filters(client, db, profile, url, user):
    response = client.get(url("/profiles/", {"name__in": [user.name, "Din Djarin"]}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    assert result["profiles"][0]["username"] == user.username


@parametrize("include,expectation", [(["skills", "team"], True), ([], False)])
@parametrize("url_params__array_format", ["comma", "repeat"])
def test_profile_list_include(client, db, expectation, include, profile, url, user):
    response = client.get(url("/profiles/", {"include": include}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    profile = result["profiles"][0]
    assert ("skills" in profile) is expectation
    assert ("team" in profile) is expectation


def test_profile_list_negated_filters(client, db, profile, url, user):
    response = client.get(url("/profiles/", {"firstName!": user.first_name}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 0


def test_profile_list_negated__icontains__filters(client, db, profile, url, user):
    response = client.get(url("/profiles/", {"firstName__icontains!": user.first_name}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 0


@parametrize("url_params__array_format", ["comma", "repeat"])
def test_profile_list_not_in_filters(client, db, profile, url, user):
    response = client.get(url("/profiles/", {"name__in!": [user.name, "Din Djarin"]}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 0


@parametrize("url_params__array_format", ["comma", "repeat"])
def test_profile_list_or_filters(client, db, profile_factory, tag_factory, url):
    tag1, tag2, tag3 = tag_factory.create_batch(3)
    profile_factory.create(tags=[tag1])
    profile_factory.create(tags=[tag2])
    profile_factory.create(tags=[tag1, tag2])
    profile_factory.create(tags=[tag3])
    profile_factory.create()
    response = client.get(url("/profiles/", {"tags__in": [tag1.pk, tag2.pk]}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 3


@patch("django.core.files.storage.FileSystemStorage.save")
def test_profile_multipart_create(mock_save, client, db, role, user):
    avatar = SimpleUploadedFile("avatar.jpg", b"", content_type="image/jpeg")
    mock_save.return_value = "avatar.jpg"
    payload = dict(avatar=avatar, role=role.pk, user=user.pk)
    response = client.post("/profiles/", payload)
    result = response.json()
    assert response.status_code == 201, result
    assert result["avatar"] == "/avatar.jpg"
    assert result["role"]["id"] == role.pk
    assert result["role"]["name"] == role.name
    assert result["user"]["username"] == user.username


@patch("django.core.files.storage.FileSystemStorage.save")
@parametrize("method", ["PATCH", "PUT"])
def test_profile_multipart_update(mock_save, client, db, method, profile, role, user):
    resume = SimpleUploadedFile("resume.pdf", b"", content_type="application/pdf")
    mock_save.return_value = "resume.pdf"
    payload = dict(resume=resume, role=role.pk, user=user.pk)
    response = client.generic(method, f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert result["resume"] == "/resume.pdf"
    assert result["role"]["id"] == role.pk
    assert result["role"]["name"] == role.name
    assert result["user"]["username"] == user.username


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_fk(client, db, method, profile, role, team):
    payload = dict(role=role.pk, team=team.pk)
    response = client.generic(method, f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert result["role"]["name"] == role.name
    assert result["team"]["name"] == team.name


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_fk_invalid_role(client, db, method, profile, role):
    response = client.generic(method, f"/profiles/{profile.pk}/", dict(role=123))
    result = response.json()
    assert response.status_code == 422, result
    assert result["message"] == "Invalid role"


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_fk_role_is_not_nullable(client, db, method, profile, role):
    response = client.generic(method, f"/profiles/{profile.pk}/", dict(role=None))
    result = response.json()
    assert response.status_code == 422, result
    assert result["message"] == "Invalid role"


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_fk_team_is_nullable(client, db, method, profile, role, team):
    response = client.generic(method, f"/profiles/{profile.pk}/", dict(team=None))
    result = response.json()
    assert response.status_code == 200, result
    assert result["team"] is None


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m(client, db, method, profile, tag):
    response = client.generic(method, f"/profiles/{profile.pk}/", dict(tags=[tag.pk]))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["tags"]) == 1
    assert result["tags"][0] == tag.name


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m_can_be_empty(client, db, method, profile, tag):
    response = client.generic(method, f"/profiles/{profile.pk}/", dict(tags=[]))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["tags"]) == 0


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m_lookup_field(client, db, method, profile, task):
    payload = dict(tasks=[task.custom_id])
    response = client.generic(method, f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["tasks"]) == 1
    assert result["tasks"][0] == task.name


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m_is_not_nullable(client, db, method, profile, tag):
    response = client.generic(method, f"/profiles/{profile.pk}/", dict(tags=None))
    result = response.json()
    assert response.status_code == 422, result
    assert "tags accepts an array, got <class 'NoneType'> None" in result["message"]


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m_must_be_pks(client, db, method, profile, tag):
    payload = dict(tags=["invalid"])
    response = client.generic(method, f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid tags" in result["message"]


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m_through(client, db, method, profile, skill):
    payload = dict(skills=[dict(id=skill.pk, rating=4)])
    response = client.generic(method, f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["skills"]) == 1
    assert result["skills"][0]["id"] == skill.pk
    assert result["skills"][0]["name"] == skill.name
    assert result["skills"][0]["rating"] == 4


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m_through_can_be_empty(client, db, method, profile, skill):
    response = client.generic(method, f"/profiles/{profile.pk}/", dict(skills=[]))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["skills"]) == 0


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m_through_is_not_nullable(client, db, method, profile, skill):
    response = client.generic(method, f"/profiles/{profile.pk}/", dict(skills=None))
    result = response.json()
    assert response.status_code == 422, result
    assert "skills accepts an array, got <class 'NoneType'> None" in result["message"]


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m_through_must_be_dicts(client, db, method, profile, skill):
    payload = dict(skills=["invalid"])
    response = client.generic(method, f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid skills" == result["message"]


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m_through_ids_must_be_pks(client, db, method, profile, skill):
    payload = dict(skills=[dict(id="invalid")])
    response = client.generic(method, f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid skills" in result["message"]


@parametrize("method", ["PATCH", "PUT"])
def test_profile_update_m2m_through_required_fields(client, db, method, profile, skill):
    payload = dict(skills=[dict(id=skill.pk)])
    response = client.generic(method, f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid skills" in result["message"]


def test_staff_detail(admin_client, profile, user):
    response = admin_client.get(f"/staff/{profile.pk}/")
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == user.username


def test_staff_detail_is_not_found_for_user(user_client, profile, user):
    response = user_client.get(f"/staff/{profile.pk}/")
    result = response.json()
    assert response.status_code == 404, result
    assert result["message"] == "Not found"


def test_staff_detail_is_unauthorized_for_guest(client, db, profile, user):
    response = client.get(f"/staff/{profile.pk}/")
    result = response.json()
    assert response.status_code == 401, result
    assert result["message"] == "Unauthorized"


def test_user_detail(client, db, user):
    response = client.get(f"/users/{user.pk}/")
    result = response.json()
    assert response.status_code == 200, result
    assert result["id"] == user.pk
    assert result["username"] == user.username


@parametrize("url_params__array_format", ["comma", "repeat"])
def test_user_detail_fields(client, db, url, user):
    response = client.get(url(f"/users/{user.pk}/", {"fields": ["id", "username"]}))
    result = response.json()
    assert response.status_code == 200, result
    assert result == dict(id=user.pk, username=user.username)


def test_user_list(client, db, user):
    response = client.get("/users/")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 1
    assert result["users"][0]["id"] == user.pk
    assert result["users"][0]["username"] == user.username
    assert result["users"][0]["email"] == user.email


@parametrize("url_params__array_format", ["comma", "repeat"])
def test_user_list_fields(client, db, url, user):
    response = client.get(url("/users/", {"fields": ["id", "username"]}))
    result = response.json()
    assert response.status_code == 200, result
    assert result["users"] == [dict(id=user.pk, username=user.username)]
    response = client.get(url("/users/", {"fields": ["id", "name"]}))
    result = response.json()
    assert response.status_code == 400, result
    assert result == dict(message="Invalid fields: OrderedSet(['name'])")


@parametrize("url_params__array_format", ["comma", "repeat"])
def test_user_list_search(client, db, url, user):
    response = client.get(url("/users/", {"q": user.email}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 1
    response = client.get(url("/users/", {"q": user.email, "search": ["id"]}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 0
    response = client.get(url("/users/", {"q": user.email, "search": ["id", "email"]}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 1
    response = client.get(url("/users/", {"q": user.email, "search": ["id", "name"]}))
    result = response.json()
    assert response.status_code == 400, result
    assert result == dict(message="Invalid fields: {'name'}")


def test_user_list_filters(client, db, url, user_factory):
    january = "2021-01-01T00:00:00Z"
    february = "2021-02-01T00:00:00Z"
    march = "2021-03-01T00:00:00Z"

    user1 = user_factory.create(date_joined=january)
    user2 = user_factory.create(date_joined=february)
    user3 = user_factory.create(email="test3@test.com", date_joined=march)

    # filter by email
    response = client.get(url("/users/", {"email": user1.email}))
    assert len(response.json()["users"]) == 1

    # filter by string
    response = client.get(url("/users/", {"q": "example.com"}))
    assert len(response.json()["users"]) == 2
    response = client.get(url("/users/", {"q": user3.email}))
    assert len(response.json()["users"]) == 1

    # filter by username -- filter ignored, username is not in filter_fields
    response = client.get(url("/users/", {"username": user2.username}))
    assert len(response.json()["users"]) == 3

    # filter by date joined
    filters = {
        "dateJoined__gte": "2021-02-01T00:00:00Z",
        "dateJoined__lte": "2021-02-15T00:00:00Z",
    }
    response = client.get(url("/users/", filters))
    assert len(response.json()["users"]) == 1
    assert response.json()["users"][0]["username"] == user2.username


def test_user_list_sort_asc(client, db, url, user_factory):
    user_factory.create(username="a")
    user_factory.create(username="b")
    response = client.get(url("/users/", {"sort": "id"}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 2
    assert result["users"][0]["username"] == "a"
    assert result["users"][1]["username"] == "b"


def test_user_list_sort_desc(client, db, url, user_factory):
    user_factory.create(username="a")
    user_factory.create(username="b")
    response = client.get(url("/users/", {"sort": "-id"}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 2
    assert result["users"][0]["username"] == "b"
    assert result["users"][1]["username"] == "a"


@parametrize("url_params__array_format", ["comma", "repeat"])
def test_user_list_multisort(client, db, now, url, user_factory):
    date_joined = now()
    user_factory.create(username="a", date_joined=date_joined)
    user_factory.create(username="b", date_joined=date_joined - timedelta(hours=1))
    user_factory.create(username="c", date_joined=date_joined)
    user_factory.create(username="d", date_joined=date_joined)
    response = client.get(url("/users/", {"sort": ["dateJoined", "-id", "x"]}))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 4
    assert result["users"][0]["username"] == "b"
    assert result["users"][1]["username"] == "d"
    assert result["users"][2]["username"] == "c"
    assert result["users"][3]["username"] == "a"


def test_user_self(client, user_client, user):
    response = client.get("/user/")
    assert response.status_code == 401

    response = user_client.get("/user/")
    assert response.status_code == 200


def test_user_unique_create_with_existing_value(client, db, user, user_factory):
    user_factory.create(username="already_taken")
    payload = dict(username="already_taken")
    response = client.post("/users/", payload)
    result = response.json()
    assert response.status_code == 422, result
    assert result["message"] == "Field username must be unique"


@parametrize("method", ["PATCH", "PUT"])
def test_user_unique_update_with_current_value(client, db, method, user, user_factory):
    payload = dict(username=user.username)
    response = client.generic(method, f"/users/{user.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == user.username


@parametrize("method", ["PATCH", "PUT"])
def test_user_unique_update_with_existing_value(client, db, method, user, user_factory):
    user_factory.create(username="already_taken")
    payload = dict(username="already_taken")
    response = client.generic(method, f"/users/{user.pk}/", payload)
    result = response.json()
    assert response.status_code == 422, result
    assert result["message"] == "Field username must be unique"


@parametrize("method", ["PATCH", "PUT"])
def test_user_update(client, db, method, user):
    payload = dict(username="testtest", email="something@example.com")
    response = client.generic(method, f"/users/{user.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == "testtest"
    assert result["email"] == "something@example.com"



from worf.serializers import Serializer, fields


class UserSerializer(Serializer):
    last_login = fields.DateTime(dump_only=True)
    date_joined = fields.DateTime(dump_only=True)

    class Meta:
        fields = [
            "id",
            "username",
            "last_login",
            "date_joined",
            "email",
        ]


class ProfileSerializer(Serializer):
    username = fields.String(attribute="user.username")
    email = fields.String(attribute="user.email")
    avatar = fields.File(lambda profile: profile.get_avatar_url())
    role = fields.Nested("RoleSerializer")
    skills = fields.Nested("SkillSerializer", attribute="ratedskill_set", many=True)
    team = fields.Nested("TeamSerializer")
    tags = fields.Pluck("TagSerializer", "name", many=True)
    tasks = fields.Pluck("TaskSerializer", "name", many=True)
    user = fields.Nested("UserSerializer")

    class Meta:
        fields = [
            "id",
            "username",
            "email",
            "phone",
            "avatar",
            "boolean",
            "integer",
            "json",
            "positive_integer",
            "slug",
            "small_integer",
            "recovery_email",
            "resume",
            "role",
            "skills",
            "team",
            "tags",
            "tasks",
            "user",
            "last_active",
            "created_at",
        ]
        writable = [
            "id",
            "email",
            "phone",
            "avatar",
            "boolean",
            "integer",
            "json",
            "positive_integer",
            "slug",
            "small_integer",
            "recovery_email",
            "resume",
            "role",
            "skills",
            "team",
            "tags",
            "tasks",
            "user",
            "last_active",
            "created_at",
        ]


class RoleSerializer(Serializer):
    class Meta:
        fields = ["id", "name"]


class SkillSerializer(Serializer):
    id = fields.Integer(attribute="skill.id")
    name = fields.String(attribute="skill.name")

    class Meta:
        fields = ["id", "name", "rating"]


class TagSerializer(Serializer):
    class Meta:
        fields = ["id", "name"]


class TaskSerializer(Serializer):
    id = fields.UUID(attribute="custom_id")

    class Meta:
        fields = ["id", "name"]


class TeamSerializer(Serializer):
    class Meta:
        fields = ["id", "name"]



from functools import wraps

import pytest
import pytest_factoryboy

fixture = pytest.fixture
raises = pytest.raises
register = pytest_factoryboy.register


@wraps(pytest.mark.kwparametrize)
def parametrize(*args, **kwargs):
    if not args:
        return pytest.mark.kwparametrize(kwargs)
    if isinstance(args[0], dict):
        return pytest.mark.kwparametrize(*args, **kwargs)
    if len(args) == 1 and not kwargs:
        return pytest.mark.kwparametrize(*args[0])
    return pytest.mark.parametrize(*args, **kwargs)



import subprocess

from worf import __version__
from worf.casing import camel_to_snake


def field_list(value):
    return [
        ".".join(map(camel_to_snake, field.split("."))) for field in string_list(value)
    ]


def get_version():
    try:
        hash = (
            subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
            .strip()
            .decode()
        )
        return f"{__version__}@{hash}"
    except:  # pragma: no cover # noqa E722 Dont crash for any reason whatsoever
        return __version__


def string_list(value):
    return value.split(",") if isinstance(value, str) else value



from tests import parametrize
from worf import exceptions


@parametrize(
    dict(e=exceptions.ActionError("test")),
    dict(e=exceptions.AuthenticationError("test")),
    dict(e=exceptions.DataConflict("test")),
    dict(e=exceptions.FieldError("test")),
    dict(e=exceptions.NamingThingsError("test")),
    dict(e=exceptions.NotFound("test")),
    dict(e=exceptions.PermissionsError("test")),
    dict(e=exceptions.WorfError("test")),
)
def test_exception(e):
    assert e.message == str(e) == "test"



from urllib.parse import urlencode

from tests import fixture, register


def pytest_configure():
    """Initialize Django settings."""
    import django
    from django.conf import settings

    settings.configure(
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": "db.sqlite3",
            }
        },
        INSTALLED_APPS=[
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "tests",
            "worf",
        ],
        MIDDLEWARE=[
            "django.contrib.sessions.middleware.SessionMiddleware",
            "django.contrib.auth.middleware.AuthenticationMiddleware",
        ],
        PASSWORD_HASHERS=["django.contrib.auth.hashers.MD5PasswordHasher"],
        ROOT_URLCONF="tests.urls",
        SECRET_KEY="secret",
        TEMPLATES=[
            {
                "BACKEND": "django.template.backends.django.DjangoTemplates",
                "APP_DIRS": True,
            },
        ],
        TIME_ZONE="UTC",
        USE_I18N=False,
        USE_L10N=False,
        USE_TZ=True,
        WORF_API_NAME="Test API",
        WORF_DEBUG=False,
    )

    django.setup()

    from tests.factories import (
        ProfileFactory,
        RoleFactory,
        SkillFactory,
        TagFactory,
        TaskFactory,
        TeamFactory,
        UserFactory,
    )

    register(ProfileFactory, "profile")
    register(RoleFactory, "role")
    register(SkillFactory, "skill")
    register(TagFactory, "tag")
    register(TaskFactory, "task")
    register(TeamFactory, "team")
    register(UserFactory, "user")


@fixture(name="admin_client")
def admin_client_fixture(db, admin_user):
    from worf.testing import ApiClient

    client = ApiClient()
    client.force_login(admin_user)
    return client


@fixture(name="client")
def client_fixture():
    from worf.testing import ApiClient

    return ApiClient()


@fixture(name="now")
def now_fixture():
    from django.utils import timezone

    return timezone.now


@fixture(name="url")
def url_fixture(url_params):
    def url(url, data={}):
        return f"{url}?{url_params(data)}" if data else url

    return url


@fixture(name="url_params")
def url_params_fixture(url_params__array_format):
    def url_params(data):
        if url_params__array_format == "comma":
            data = {
                key: ",".join(map(str, value)) if isinstance(value, list) else value
                for key, value in data.items()
            }
        return urlencode(data, True)

    return url_params


@fixture(name="url_params__array_format")
def url_params__array_format_fixture():
    return "repeat"


@fixture(name="user_client")
def user_client_fixture(db, user):
    from worf.testing import ApiClient

    client = ApiClient()
    client.force_login(user)
    return client



import marshmallow
from marshmallow.decorators import *  # noqa: F401, F403

from django.core.exceptions import ImproperlyConfigured
from django.db.models.fields.files import FieldFile

from worf import fields
from worf.casing import snake_to_camel
from worf.conf import settings
from worf.exceptions import FieldError
from worf.shortcuts import field_list


class SerializeModels:
    include_fields = {}
    serializer = None
    staff_serializer = None

    def get_serializer(self):
        serializer = self.serializer

        if self.staff_serializer and self.request.user.is_staff:  # pragma: no cover
            serializer = self.staff_serializer

        if not serializer:  # pragma: no cover
            msg = f"{self.__class__.__name__}.get_serializer() did not return a serializer"
            raise ImproperlyConfigured(msg)

        return serializer(**self.get_serializer_kwargs())

    def get_serializer_context(self):
        return {}

    def get_serializer_kwargs(self):
        context = dict(request=self.request, **self.get_serializer_context())
        only = set(field_list(self.bundle.get("fields", [])))
        include = field_list(self.bundle.get("include", []))
        exclude = set(self.include_fields.keys()) - set(include)
        return dict(context=context, only=only, exclude=exclude)

    def load_serializer(self):
        try:
            return self.get_serializer()
        except ValueError as e:
            if str(e).startswith("Invalid fields"):
                invalid_fields = str(e).partition(": ")[2].strip(".")
                raise FieldError(f"Invalid fields: {invalid_fields}")
            raise e  # pragma: no cover

    def serialize(self):
        return self.load_serializer().dump(self.get_instance())


class SerializerOptions(marshmallow.SchemaOpts):
    def __init__(self, meta, **kwargs):
        super().__init__(meta, **kwargs)

        defaults = settings.WORF_SERIALIZER_DEFAULT_OPTIONS
        defaults["ordered"] = defaults.get("ordered", True)

        for key, value in defaults.items():
            setattr(self, key, getattr(meta, key, value))

        fields = getattr(meta, "fields", [])
        writable = getattr(meta, "writable", [])

        if writable:
            self.dump_only = list(set(fields) - set(writable))


class Serializer(marshmallow.Schema):
    """
    Serializers are basically marshmallow schemas with extra handling for
    Django models/managers, some tweaks to field naming/ordering and the option
    to supply some meta defaults via Django settings.

    The read/write methods are relics of our interim serialization strategy
    they're analogous with marshmallow dump/load, but a bit less flexible.

    Marshmallow is primarily used for field filtering right now, worf coerces
    values itself, but we might delegated more to marshmallow later.
    """

    OPTIONS_CLASS = SerializerOptions

    TYPE_MAPPING = {
        **marshmallow.Schema.TYPE_MAPPING,
        FieldFile: fields.File,
    }

    def __init__(self, only=(), *args, **kwargs):
        super().__init__(only=only or None, *args, **kwargs)

    def __call__(self, **kwargs):
        only = self.only
        if self.only and kwargs.get("only"):
            invalid_fields = set(kwargs.get("only")) - self.only
            if invalid_fields:
                raise FieldError(f"Invalid fields: {invalid_fields}")
            only = set(kwargs.get("only"))
        elif kwargs.get("only"):
            only = kwargs.get("only")

        return self.__class__(
            context=kwargs.get("context", self.context),
            dump_only=kwargs.get("dump_only", self.dump_only),
            exclude=set(self.exclude or []) | set(kwargs.get("exclude") or []),
            load_only=kwargs.get("load_only", self.load_only),
            many=kwargs.get("many", self.many),
            only=only,
            partial=kwargs.get("partial", self.partial),
            unknown=kwargs.get("unknown", self.unknown),
        )

    def __repr__(self):
        name = self.__class__.__name__
        kwargs = dict(
            exclude=sorted(self.exclude or []),
            many=self.many,
            only=sorted(self.only or []),
        )
        return f"<{name}({', '.join(f'{k}={v}' for k, v in kwargs.items() if v)})>"

    @property
    def dict_class(self):
        return dict

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = snake_to_camel(field_obj.data_key or field_name)

    class Meta:
        ordered = True



def test_browsable_api(client, db, profile, user):
    profile_url = f"/profiles/{profile.pk}/"
    response = client.get(profile_url, HTTP_ACCEPT="text/html,image/png")

    content = response.content.decode("UTF-8")

    assert response.status_code == 200, content
    assert "Test API" in content
    assert "/api/" in content
    assert "HTTP" in content
    assert "200" in content
    assert "Content-Type" in content
    assert "application/json" in content
    assert user.username in content

    assert client.get(profile_url, HTTP_ACCEPT="").json()
    assert client.get(profile_url, HTTP_ACCEPT="application/json").json()
    assert client.get(profile_url, HTTP_ACCEPT="text/plain").json()
    assert client.get(f"{profile_url}?format=json", HTTP_ACCEPT="text/html").json()



from django.http import HttpResponse, JsonResponse
from django.template.response import TemplateResponse

from worf.conf import settings


def browsable_response(request, response, status_code=200):
    template = "worf/api.html"
    context = dict(
        content=response.content.decode("utf-8"),
        response=response,
        settings=settings,
    )

    response = TemplateResponse(request, template, context=context)
    response.status_code = status_code
    response.render()

    return response


def render_response(request, data, status_code=200):
    is_browsable = (
        settings.WORF_BROWSABLE_API
        and "text/html" in request.headers.get("Accept", "")
        and request.GET.get("format") != "json"
    )

    response = (
        JsonResponse(data, json_dumps_params=dict(indent=2 if is_browsable else 0))
        if data != ""
        else HttpResponse()
    )

    response.status_code = status_code

    if is_browsable:
        response = browsable_response(request, response, status_code)

    return response



from worf.assigns import AssignAttributes
from worf.lookups import FindInstance
from worf.views.base import AbstractBaseAPI


class UpdateAPI(AssignAttributes, FindInstance, AbstractBaseAPI):
    update_serializer = None

    def get_serializer(self):
        if self.update_serializer and self.request.method in ("PATCH", "PUT"):
            return self.update_serializer(**self.get_serializer_kwargs())
        return super().get_serializer()

    def patch(self, *args, **kwargs):
        self.update(*args, **kwargs)
        return self.render_to_response()

    def put(self, *args, **kwargs):
        self.update(*args, **kwargs)
        return self.render_to_response()

    def update(self, *args, **kwargs):
        instance = self.get_instance()
        self.validate()
        self.save(instance, self.bundle)
        instance.refresh_from_db()
        return instance



from tests.serializers import (
    ProfileSerializer,
    RoleSerializer,
    SkillSerializer,
    TagSerializer,
    TaskSerializer,
    TeamSerializer,
    UserSerializer,
)


def test_profile_serializer():
    assert f"{ProfileSerializer()}"


def test_role_serializer():
    assert f"{RoleSerializer()}"


def test_skill_serializer():
    assert f"{SkillSerializer()}"


def test_tag_serializer():
    assert f"{TagSerializer()}"


def test_task_serializer():
    assert f"{TaskSerializer()}"


def test_team_serializer():
    assert f"{TeamSerializer()}"


def test_user_serializer():
    assert f"{UserSerializer()}"



import operator
from functools import reduce

from django.core.exceptions import EmptyResultSet, ImproperlyConfigured
from django.core.paginator import EmptyPage, Paginator
from django.db.models import F, OrderBy, Prefetch, Q

from worf.casing import camel_to_snake
from worf.conf import settings
from worf.exceptions import FieldError
from worf.filters import apply_filterset, generate_filterset
from worf.shortcuts import field_list, string_list
from worf.views.base import AbstractBaseAPI
from worf.views.create import CreateAPI


class ListAPI(AbstractBaseAPI):
    lookup_url_kwarg = "id"  # default incase lookup_field is set
    ordering = []
    filter_fields = []
    include_fields = {}
    search_fields = []
    sort_fields = []
    queryset = None
    filter_set = None
    count = 0
    page_num = 1
    per_page = 25
    max_per_page = None
    num_pages = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        codepath = self.codepath

        if not isinstance(self.ordering, list):  # pragma: no cover
            raise ImproperlyConfigured(f"{codepath}.ordering must be a list")

        if not isinstance(self.filter_fields, list):  # pragma: no cover
            raise ImproperlyConfigured(f"{codepath}.filter_fields must be a list")

        if not isinstance(self.include_fields, dict):  # pragma: no cover
            raise ImproperlyConfigured(f"{codepath}.include_fields must be a dict")

        if not isinstance(self.search_fields, list):  # pragma: no cover
            raise ImproperlyConfigured(f"{codepath}.search_fields must be a list")

        if not isinstance(self.sort_fields, list):  # pragma: no cover
            raise ImproperlyConfigured(f"{codepath}.sort_fields must be a list")

        # generate a default filterset if a custom one was not provided
        if self.filter_set is None:
            self.filter_set = generate_filterset(self.model, self.queryset)

    def get(self, *args, **kwargs):
        return self.render_to_response()

    def set_base_lookup_kwargs(self):
        if hasattr(self, "lookup_field") and hasattr(self, "lookup_url_kwarg"):
            self.lookup_kwargs[self.lookup_field] = self.kwargs[self.lookup_url_kwarg]

    def set_search_lookup_kwargs(self):
        """
        Set generic lookup kwargs based on q and additional GET params.

        For more advanced search use cases, override this method and pass GET
        with any remaining params you want to use classic django filters for.
        """
        if not self.filter_fields and not self.search_fields:  # pragma: no cover
            return

        # Whatever is not q or page as a querystring param will
        # be used for key-value search.
        query = self.bundle.pop("q", "").strip()

        self.bundle.pop("page", None)
        self.bundle.pop("p", None)

        if query:
            search_fields = self.search_fields

            if self.bundle.get("search", []):
                search_fields = field_list(self.bundle["search"])
                invalid_fields = set(search_fields) - set(self.search_fields)
                if invalid_fields:
                    raise FieldError(f"Invalid fields: {invalid_fields}")

            if search_fields:
                search_icontains = (
                    Q(**{f"{search_field}__icontains": query})
                    for search_field in search_fields
                )
                self.search_query = reduce(operator.or_, search_icontains)

        if not self.filter_fields or not self.bundle:  # pragma: no cover
            return

        for key in self.bundle.keys():
            filter_key = key[:-1] if key.endswith("!") else key

            if filter_key not in self.filter_fields:
                continue

            value = self.bundle[key]

            # support passing `in` and `range` as lists
            if isinstance(value, list):
                if filter_key.endswith("__in") or filter_key.endswith("__range"):
                    value = ",".join(str(item) for item in value)

            self.lookup_kwargs[key] = value

    def get_queryset(self):
        if self.queryset is None:
            return self.model.objects.all()
        return self.queryset.all()

    def get_processed_queryset(self):
        self.lookup_kwargs = {}
        self.search_query = Q()

        self.set_base_lookup_kwargs()
        self.set_search_lookup_kwargs()

        lookups = self.lookup_kwargs.items()
        filterset_kwargs = {k: v for k, v in lookups if not isinstance(v, list)}
        list_kwargs = {k: v for k, v in lookups if isinstance(v, list)}
        ordering = self.get_ordering()

        queryset = (
            apply_filterset(self.filter_set, self.get_queryset(), filterset_kwargs)
            .filter(self.search_query)
            .distinct()
        )

        for key, value in list_kwargs.items():
            for item in value:
                queryset = (
                    queryset.exclude(**{key.rstrip("!"): item})
                    if key.endswith("!")
                    else queryset.filter(**{key: item})
                )

        if self.include_fields and self.bundle.get("include"):
            include = field_list(self.bundle["include"])
            for item in set(self.include_fields.keys()) & set(include):
                if isinstance(self.include_fields[item], Prefetch):
                    queryset = queryset.prefetch_related(self.include_fields[item])
                elif isinstance(self.include_fields[item], str):
                    queryset = queryset.select_related(self.include_fields[item])

        if ordering:
            queryset = queryset.order_by(*ordering)

        return queryset

    def get_ordering(self):
        ordering = []

        for sort in string_list(self.bundle.get("sort", [])):
            field = "__".join(map(camel_to_snake, sort.lstrip("-").split(".")))
            if field not in self.sort_fields:
                continue
            ordering.append(self.get_sort_field(field, descending=sort[0] == "-"))

        return ordering or self.ordering

    def get_sort_field(self, field, descending=False):
        return OrderBy(F(field), descending=descending)

    def paginated_results(self):
        queryset = self.get_processed_queryset()
        request = self.request

        if settings.WORF_DEBUG:  # pragma: no cover
            try:
                self.query = str(queryset.query)
            except EmptyResultSet:
                self.query = None

        default_per_page = getattr(self, "results_per_page", self.per_page)
        per_page = max(int(request.GET.get("perPage") or default_per_page), 1)
        max_per_page = self.max_per_page or default_per_page
        paginator = Paginator(queryset, min(per_page, max_per_page))

        self.page_num = int(request.GET.get("page") or request.GET.get("p") or 1)
        if self.page_num < 1:
            self.page_num = 1

        self.num_pages = paginator.num_pages
        self.count = paginator.count

        try:
            return paginator.page(self.page_num)
        except EmptyPage:
            return []

    def serialize(self):
        serializer = self.load_serializer()

        payload = {str(self.name): serializer(many=True).dump(self.paginated_results())}

        if self.per_page:
            payload["pagination"] = dict(
                count=self.count,
                pages=self.num_pages,
                page=self.page_num,
            )

        if settings.WORF_DEBUG:  # pragma: no cover
            payload["debug"] = {
                "bundle": self.bundle,
                "lookup_kwargs": getattr(self, "lookup_kwargs", {}),
                "query": self.query,
                "search_query": str(self.search_query),
                "serializer": str(serializer).strip("<>"),
            }

        return payload


class ListCreateAPI(CreateAPI, ListAPI):
    pass



from worf.assigns import AssignAttributes
from worf.views.base import AbstractBaseAPI


class CreateAPI(AssignAttributes, AbstractBaseAPI):
    create_serializer = None

    def create(self, *args, **kwargs):
        self.instance = self.new_instance()
        self.validate()
        self.save(self.instance, self.bundle)
        self.instance.refresh_from_db()
        return self.instance

    def get_serializer(self):
        if self.create_serializer and self.request.method == "POST":
            return self.create_serializer(**self.get_serializer_kwargs())
        return super().get_serializer()

    def new_instance(self):
        return self.model()

    def post(self, *args, **kwargs):
        instance = self.create(*args, **kwargs)
        result = self.load_serializer().dump(instance) if instance else ""
        return self.render_to_response(result, 201)



class FindInstance:
    lookup_field = "id"
    lookup_url_kwarg = "id"
    queryset = None

    def get_instance(self):
        if not hasattr(self, "instance"):
            self.lookup_kwargs = {self.lookup_field: self.kwargs[self.lookup_url_kwarg]}
            self.instance = self.get_queryset().get(**self.lookup_kwargs)

        return self.instance

    def get_queryset(self):
        if self.queryset is None:
            return self.model.objects.all()
        return self.queryset.all()



import pytest

from django.contrib.auth.models import AnonymousUser, User

from worf.exceptions import AuthenticationError, NotFound
from worf.permissions import Authenticated, PublicEndpoint, Staff


def test_authenticated(db, rf):
    permission = Authenticated()
    request = rf.get("/")
    request.user = AnonymousUser()

    with pytest.raises(AuthenticationError):
        assert permission(request) is None

    request.user = User.objects.create(username="test", password="test")
    assert permission(request) is None


def test_public_endpoint(db, rf):
    permission = PublicEndpoint()
    request = rf.get("/")
    request.user = AnonymousUser()
    assert permission(request) is None
    request.user = User.objects.create(username="test", password="test")
    assert permission(request) is None


def test_staff(db, rf):
    permission = Staff()
    request = rf.get("/")
    request.user = AnonymousUser()

    with pytest.raises(NotFound):
        assert permission(request) is None

    request.user = User.objects.create(username="test", password="test")

    with pytest.raises(NotFound):
        assert permission(request) is None

    request.user.is_staff = True
    request.user.save()
    permission(request)



from worf.exceptions import ActionError
from worf.lookups import FindInstance
from worf.views.base import AbstractBaseAPI


class ActionAPI(FindInstance, AbstractBaseAPI):
    actions = []

    def dispatch(self, request, *args, **kwargs):
        if "action" in kwargs and request.method != "PUT":
            return self.http_method_not_allowed(request, *args, **kwargs)
        return super().dispatch(request, *args, **kwargs)

    def perform_action(self, request, *args, **kwargs):
        action = kwargs["action"].replace("-", "_")
        if action not in self.actions:
            raise ActionError(f"Invalid action: {kwargs['action']}")
        instance = self.get_instance()
        if hasattr(self, action):
            return getattr(self, action)(request, **self.bundle, user=request.user)
        try:
            getattr(instance, action)(**self.bundle, user=request.user)
        except TypeError as e:
            if "unexpected keyword argument 'user'" not in str(e):
                raise ActionError(f"Invalid arguments: {e}")
            getattr(instance, action)(**self.bundle)
        return instance

    def put(self, request, *args, **kwargs):
        if "action" in kwargs:
            self.perform_action(request, *args, **kwargs)
            return self.render_to_response()
        return super().put(request, *args, **kwargs)



from tests import parametrize


def test_action_model_func(user_client, profile):
    response = user_client.put(f"/profiles/{profile.pk}/unsubscribe/")
    result = response.json()
    assert response.status_code == 200, result
    profile.refresh_from_db()
    assert profile.is_subscribed is False
    assert profile.subscribed_at is None
    assert profile.subscribed_by is None


def test_action_model_func_with_user_argument(user_client, profile, user):
    response = user_client.put(f"/profiles/{profile.pk}/subscribe/")
    result = response.json()
    assert response.status_code == 200, result
    profile.refresh_from_db()
    assert profile.is_subscribed is True
    assert profile.subscribed_at is not None
    assert profile.subscribed_by == user


def test_action_view_func(user_client, profile, user):
    response = user_client.put(f"/profiles/{profile.pk}/resubscribe/")
    result = response.json()
    assert response.status_code == 200, result
    profile.refresh_from_db()
    assert profile.is_subscribed is True
    assert profile.subscribed_at is not None
    assert profile.subscribed_by == user


def test_invalid_action(user_client, profile):
    response = user_client.put(f"/profiles/{profile.pk}/invalid-action/")
    result = response.json()
    assert response.status_code == 400, result
    assert result["message"] == "Invalid action: invalid-action"


def test_invalid_arguments(user_client, profile):
    kwargs = dict(text="I love newsletters")
    response = user_client.put(f"/profiles/{profile.pk}/subscribe/", kwargs)
    result = response.json()
    assert response.status_code == 400, result
    assert result["message"].startswith("Invalid arguments:")
    assert result["message"].endswith("unexpected keyword argument 'text'")


@parametrize("method", ["GET", "DELETE", "PATCH", "POST"])
def test_invalid_method(user_client, method, profile):
    response = user_client.generic(method, f"/profiles/{profile.pk}/subscribe/")
    assert response.status_code == 405



from marshmallow import fields, utils
from marshmallow.exceptions import ValidationError
from marshmallow.fields import *  # noqa: F401, F403

from django.db.models import Manager


class File(fields.Field):
    _CHECK_ATTRIBUTE = False

    def __init__(self, serialize=None, deserialize=None, **kwargs):
        super().__init__(**kwargs)
        self.serialize_func = serialize and utils.callable_or_raise(serialize)
        self.deserialize_func = deserialize and utils.callable_or_raise(deserialize)

    def _serialize(self, value, attr, obj, **kwargs):
        if self.serialize_func:
            return self._call_or_raise(self.serialize_func, obj, attr)
        return value.url if value.name else None

    # Worf serializers are not really used for deserialization yet, so no cover

    def _deserialize(self, value, attr, data, **kwargs):  # pragma: no cover
        if self.deserialize_func:
            return self._call_or_raise(self.deserialize_func, value, attr)
        return value

    def _call_or_raise(self, func, value, attr):
        if len(utils.get_func_args(func)) > 1:  # pragma: no cover
            if self.parent.context is None:
                msg = f"No context available for Function field {attr!r}"
                raise ValidationError(msg)
            return func(value, self.parent.context)
        return func(value)


class Nested(fields.Nested):
    def _serialize(self, nested_obj, attr, obj, **kwargs):
        if isinstance(nested_obj, Manager):
            nested_obj = nested_obj.all()
        return super()._serialize(nested_obj, attr, obj, **kwargs)


class Pluck(fields.Pluck):
    def _serialize(self, nested_obj, attr, obj, **kwargs):
        if isinstance(nested_obj, Manager):
            nested_obj = nested_obj.all()
        return super()._serialize(nested_obj, attr, obj, **kwargs)



from worf.exceptions import AuthenticationError, NotFound


class Authenticated:
    def __call__(self, request, **kwargs):
        if not request.user.is_authenticated:
            raise AuthenticationError()


class PublicEndpoint:
    def __call__(self, request, **kwargs):
        pass


class Staff:
    def __call__(self, request, **kwargs):
        if not request.user.is_authenticated or not request.user.is_staff:
            raise NotFound()



from functools import partialmethod

from django.core.files import File
from django.test.client import MULTIPART_CONTENT, Client

JSON_CONTENT = "application/json"


class ApiClient(Client):
    def generic(self, method, path, data=None, content_type=None, **kwargs):
        content_type = content_type or guess_content_type(data)
        data = self._encode_multipart(data, content_type)
        return super().generic(method, path, data, content_type, **kwargs)

    delete = partialmethod(generic, "DELETE")
    patch = partialmethod(generic, "PATCH")
    post = partialmethod(generic, "POST")
    put = partialmethod(generic, "PUT")

    def _encode_multipart(self, data, content_type):
        post_data = self._encode_json({} if data is None else data, content_type)
        return self._encode_data(post_data, content_type)


def guess_content_type(data):
    return MULTIPART_CONTENT if should_multipart(data) else JSON_CONTENT


def has_files(data):
    return any(isinstance(value, File) for value in data.values())


def should_multipart(data):
    return isinstance(data, dict) and has_files(data)



from urllib.parse import urlencode

from url_filter.exceptions import SkipFilter
from url_filter.filtersets import ModelFilterSet

from django.db.models.fields.related import ForeignObjectRel, RelatedField
from django.http import QueryDict


class AnnotatedModelFilterSet(ModelFilterSet):
    def get_filters(self):
        filters = super().get_filters()

        if self.queryset is not None:
            state = self._build_state()

            for name in self.queryset.query.annotations.keys():
                if name in self.Meta.exclude or name in filters:  # pragma: no cover
                    continue

                try:
                    annotation_filter = self._build_annotation_filter(name, state)
                except SkipFilter:  # pragma: no cover
                    continue

                if annotation_filter is not None:
                    filters[name] = annotation_filter

        return filters

    def _build_annotation_filter(self, name, state):
        field = self.queryset.query.annotations.get(name).output_field

        if isinstance(field, RelatedField):  # pragma: no cover
            if not self.Meta.allow_related:
                raise SkipFilter
            return self._build_filterset_from_related_field(name, field)
        elif isinstance(field, ForeignObjectRel):  # pragma: no cover
            if not self.Meta.allow_related_reverse:
                raise SkipFilter
            return self._build_filterset_from_reverse_field(name, field)

        return self._build_filter_from_field(name, field)


def generate_filterset(model, queryset):
    return type(
        f"{model.__name__}FilterSet",
        (AnnotatedModelFilterSet,),
        dict(Meta=type("Meta", (), dict(model=model, queryset=queryset))),
    )


def apply_filterset(filter_set, queryset, lookup_kwargs):
    data = QueryDict(urlencode(lookup_kwargs, True))

    return filter_set(data=data, queryset=queryset).filter()



from django.core.exceptions import ValidationError
from django.db import models
from django.db.utils import IntegrityError


class AssignAttributes:
    def save(self, instance, bundle):
        items = [
            (key, getattr(self.model, key), value) for key, value in bundle.items()
        ]

        for key, attr, value in items:
            if isinstance(value, models.Model):
                setattr(instance, key, value)
                continue

            if isinstance(attr.field, models.ForeignKey):
                self.set_foreign_key(instance, key, value)
                continue

            if isinstance(attr.field, models.ManyToManyField):
                continue

            setattr(instance, key, value)

        instance.save()

        for key, attr, value in items:
            if isinstance(attr.field, models.ManyToManyField):
                has_through = not attr.through._meta.auto_created

                if has_through and any(isinstance(item, dict) for item in value):
                    self.set_many_to_many_with_through(instance, key, value)
                    continue

                self.set_many_to_many(instance, key, value)

    def resolve_relation(self, key, value):
        related_model = getattr(self.model, key).field.related_model
        lookup_field = getattr(getattr(related_model, "Api", ""), "lookup_field", "pk")
        try:
            value = related_model.objects.get(**{lookup_field: value})
        except related_model.DoesNotExist as e:
            raise ValueError from e
        return value

    def set_foreign_key(self, instance, key, value):
        try:
            value = self.resolve_relation(key, value) if value is not None else None
        except ValueError as e:
            raise ValidationError(f"Invalid {self.keymap[key]}") from e
        setattr(instance, key, value)

    def set_many_to_many(self, instance, key, value):
        related_manager = getattr(instance, key)
        related_model = related_manager.model
        lookup_field = getattr(getattr(related_model, "Api", ""), "lookup_field", "pk")
        try:
            if lookup_field != "pk":
                items = related_model.objects.filter(**{f"{lookup_field}__in": value})
                assert len(items) == len(value)
                value = items
            related_manager.set(value)
        except (AssertionError, IntegrityError, ValueError) as e:
            raise ValidationError(f"Invalid {self.keymap[key]}") from e

    def set_many_to_many_with_through(self, instance, key, value):
        try:
            attr = getattr(self.model, key)
            through_model = attr.through
            model_name = self.model._meta.model_name
            target_key = attr.field.m2m_target_field_name()
            relation_name = attr.field.m2m_reverse_field_name()

            def pivot_data(item):
                return {k: v for k, v in item.items() if k != target_key}

            def target_lookup(item):
                return {relation_name: self.resolve_relation(key, item[target_key])}

            def save_through(defaults, **kwargs):
                return through_model.objects.update_or_create(defaults, **kwargs)[0]

            model_lookup = {model_name: instance}

            items = [
                save_through(pivot_data(item), **model_lookup, **target_lookup(item)).pk
                for item in value
            ]

            through_model.objects.filter(**model_lookup).exclude(pk__in=items).delete()
        except (AttributeError, IntegrityError, ValueError) as e:
            raise ValidationError(f"Invalid {self.keymap[key]}") from e

    def validate(self):
        instance = self.get_instance()

        for key in self.bundle.keys():
            self.validate_bundle(key)

            field = self.model._meta.get_field(key)

            if self.bundle[key] is None and not field.null:
                raise ValidationError(f"Invalid {self.keymap[key]}")

            if field.unique:
                other_records = self.model.objects.exclude(pk=instance.pk)

                if other_records.filter(**{key: self.bundle[key]}).exists():
                    raise ValidationError(f"Field {self.keymap[key]} must be unique")



from uuid import uuid4

import pytest

from django.core.exceptions import ValidationError

uuid = uuid4()
email = "something@example.com"
phone = "(555) 555-5555"


@pytest.fixture(name="profile_view")
def profile_view_fixture(db, now, profile_factory, rf):
    from tests.views import ProfileDetail

    profile_factory.create(email=email, phone=phone)
    view = ProfileDetail()
    view.set_bundle(
        dict(
            id=str(uuid),
            email=email,
            phone=phone,
            boolean=True,
            integer=123,
            json=dict(something=True),
            positive_integer=123,
            slug="something",
            small_integer=123,
            recovery_email=email,
            recovery_phone=phone,
            last_active=now().date().isoformat(),
            created_at=now().isoformat(),
        )
    )
    view.request = rf.patch(f"/{uuid}/")
    view.kwargs = dict(id=str(uuid))
    return view


def test_strip_fields(profile_view):
    profile_view.set_bundle(dict(slug="te\x00st "))
    profile_view.validate_bundle("slug") == "test"

    profile_view.secure_fields = ["slug"]

    profile_view.set_bundle(dict(slug="te\x00st "))
    profile_view.validate_bundle("slug") == "te\x00st "


def test_validate_bundle(profile_view):
    profile_view.validate_bundle("id")
    profile_view.validate_bundle("email")
    profile_view.validate_bundle("phone")
    profile_view.validate_bundle("boolean")
    profile_view.validate_bundle("integer")
    profile_view.validate_bundle("json")
    profile_view.validate_bundle("positive_integer")
    profile_view.validate_bundle("slug")
    profile_view.validate_bundle("small_integer")
    profile_view.validate_bundle("recovery_email")
    profile_view.validate_bundle("last_active")
    profile_view.validate_bundle("created_at")


def test_validate_bundle_accepts_blanks(profile_view):
    profile_view.set_bundle(dict(recovery_email=""))
    profile_view.validate_bundle("recovery_email")


def test_validate_bundle_accepts_nulls(profile_view):
    profile_view.set_bundle(dict(recovery_email=None))
    profile_view.validate_bundle("recovery_email")


def test_validate_bundle_raises_invalid_booleans(profile_view):
    profile_view.set_bundle(dict(boolean="nooo"))

    with pytest.raises(ValidationError) as e:
        profile_view.validate_bundle("boolean")

    assert "Field boolean accepts a boolean, got nooo" in str(e.value)


def test_validate_bundle_raises_invalid_fields(profile_view):
    profile_view.set_bundle(dict(invalid_field=email))

    with pytest.raises(ValidationError) as e:
        profile_view.validate_bundle("invalid_field")

    assert "invalid_field is not editable" in str(e.value)


def test_validate_bundle_raises_read_only_fields(profile_view):
    with pytest.raises(ValidationError) as e:
        profile_view.validate_bundle("recovery_phone")

    assert "recovery_phone is not editable" in str(e.value)


def test_validate_uuid_accepts_str(profile_view):
    assert profile_view.validate_uuid(str(uuid)) == uuid


def test_validate_uuid_accepts_uuid(profile_view):
    assert profile_view.validate_uuid(uuid) == uuid


def test_validate_uuid_raises_error(profile_view):
    with pytest.raises(ValidationError):
        profile_view.validate_uuid("not-a-uuid")


def test_validate_email_passes(profile_view):
    assert profile_view.validate_email(email) == email


def test_validate_email_raises_error(profile_view):
    with pytest.raises(ValidationError):
        profile_view.validate_email("fake.example@com")


def test_validate_custom_field_passes(profile_view):
    assert profile_view.validate_phone(phone) == "+5555555555"


def test_validate_custom_field_raises_error(profile_view):
    with pytest.raises(ValidationError):
        profile_view.validate_phone("invalid number")



from django.conf import settings

DATA_UPLOAD_MAX_MEMORY_SIZE = getattr(settings, "DATA_UPLOAD_MAX_MEMORY_SIZE")

WORF_API_NAME = getattr(settings, "WORF_API_NAME", "Worf API")
WORF_API_ROOT = getattr(settings, "WORF_API_ROOT", "/api/")

WORF_BROWSABLE_API = getattr(settings, "WORF_BROWSABLE_API", True)

WORF_DEBUG = getattr(settings, "WORF_DEBUG", settings.DEBUG)

WORF_SERIALIZER_DEFAULT_OPTIONS = getattr(
    settings, "WORF_SERIALIZER_DEFAULT_OPTIONS", {}
)



from django.urls import path

from tests import views

urlpatterns = [
    path("profiles/", views.ProfileList.as_view()),
    path("profiles/<uuid:id>/", views.ProfileDetail.as_view()),
    path("profiles/<uuid:id>/<str:action>/", views.ProfileDetail.as_view()),
    path("staff/<uuid:id>/", views.StaffDetail.as_view()),
    path("user/", views.UserSelf.as_view()),
    path("users/", views.UserList.as_view()),
    path("users/<int:id>/", views.UserDetail.as_view()),
]



from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.db.models import F, Prefetch, Value
from django.db.models.functions import Concat

from tests.models import Profile
from tests.serializers import ProfileSerializer, UserSerializer
from worf.exceptions import AuthenticationError
from worf.permissions import Authenticated, PublicEndpoint, Staff
from worf.views import ActionAPI, CreateAPI, DeleteAPI, DetailAPI, ListAPI, UpdateAPI


class ProfileList(CreateAPI, ListAPI):
    model = Profile
    queryset = Profile.objects.annotate(
        first_name=F("user__first_name"),
        last_name=F("user__last_name"),
        name=Concat("first_name", Value(" "), "last_name"),
        date_joined=F("user__date_joined"),
    )
    ordering = ["pk"]
    serializer = ProfileSerializer
    permissions = [PublicEndpoint]
    search_fields = []
    filter_fields = [
        "first_name",
        "first_name__icontains",
        "last_name",
        "name",
        "name__icontains",
        "name__in",
        "date_joined__gte",
        "tags",
        "tags__in",
    ]
    include_fields = {
        "skills": Prefetch("skills"),
        "team": "team",
    }


class ProfileDetail(ActionAPI, DeleteAPI, UpdateAPI, DetailAPI):
    model = Profile
    serializer = ProfileSerializer
    permissions = [PublicEndpoint]
    actions = [
        "resubscribe",
        "subscribe",
        "unsubscribe",
    ]

    def resubscribe(self, request, *args, **kwargs):
        profile = self.get_instance()
        profile.subscribe(user=request.user)
        return profile

    def validate_phone(self, value):
        try:
            assert value == "(555) 555-5555"
        except AssertionError:
            raise ValidationError("Field phone is not a valid phone number")
        return "+5555555555"


class StaffDetail(ProfileDetail):
    permissions = [Authenticated, Staff]


class UserList(CreateAPI, ListAPI):
    model = User
    ordering = ["pk"]
    serializer = UserSerializer(
        only=[
            "id",
            "username",
            "date_joined",
            "email",
        ]
    )
    permissions = [PublicEndpoint]
    filter_fields = [
        "id",
        "email",
        "date_joined__gte",
        "date_joined__lte",
    ]
    search_fields = [
        "id",
        "email",
        "username",
    ]
    sort_fields = [
        "id",
        "date_joined",
    ]


class UserDetail(UpdateAPI, DetailAPI):
    model = User
    serializer = UserSerializer(exclude=["date_joined"])
    permissions = [PublicEndpoint]


class UserSelf(DetailAPI):
    model = User
    serializer = UserSerializer
    permissions = [PublicEndpoint]

    def get_instance(self):
        if not self.request.user.is_authenticated:
            raise AuthenticationError("Log in with your username and password")
        return self.request.user



from worf.lookups import FindInstance
from worf.views.base import AbstractBaseAPI
from worf.views.update import UpdateAPI


class DetailAPI(FindInstance, AbstractBaseAPI):
    def get(self, *args, **kwargs):
        return self.render_to_response()


class DetailUpdateAPI(UpdateAPI, DetailAPI):
    pass



import json
import warnings
from io import BytesIO
from urllib.parse import parse_qs

from django.core.exceptions import (
    ImproperlyConfigured,
    ObjectDoesNotExist,
    RequestDataTooBig,
    ValidationError,
)
from django.template.defaultfilters import filesizeformat
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.cache import never_cache

from worf.casing import camel_to_snake, snake_to_camel
from worf.conf import settings
from worf.exceptions import (
    ActionError,
    AuthenticationError,
    DataConflict,
    FieldError,
    NotFound,
    PermissionsError,
    WorfError,
)
from worf.renderers import render_response
from worf.serializers import SerializeModels
from worf.validators import ValidateFields


@method_decorator(never_cache, name="dispatch")
class APIResponse(View):
    def __init__(self, *args, **kwargs):
        self.codepath = f"{self.__module__}.{self.__class__.__name__}"
        return super().__init__(*args, **kwargs)

    def serialize(self):
        raise NotImplementedError

    def render_to_response(self, data=None, status_code=200):
        if data is None:
            data = self.serialize()

        if data is None:
            msg = f"{self.codepath} did not pass an object to "
            msg += "render_to_response, nor did its serializer method"
            raise ImproperlyConfigured(msg)

        return render_response(self.request, data, status_code)


class AbstractBaseAPI(SerializeModels, ValidateFields, APIResponse):
    model = None
    permissions = []
    payload_key = None

    def __init__(self, *args, **kwargs):
        self.codepath = f"{self.__module__}.{self.__class__.__name__}"

        if self.model is None:  # pragma: no cover
            raise ImproperlyConfigured(f"Model is not set on {self.codepath}")

        if not isinstance(self.permissions, list):  # pragma: no cover
            raise ImproperlyConfigured(
                f"{self.codepath}.permissions must be type: list"
            )

        for method in ["post", "patch", "put", "delete"]:
            if method in dir(self) and not len(self.permissions):
                warnings.warn(
                    "\n{} method allowed on {} without permissions.".format(
                        method.upper(),
                        self.codepath,
                    ),
                )

        super().__init__(*args, **kwargs)

    @property
    def name(self):
        if isinstance(self.payload_key, str):
            return self.payload_key
        verbose_name_plural = self.model._meta.verbose_name_plural
        return snake_to_camel(verbose_name_plural.replace(" ", "_").lower())

    def dispatch(self, request, *args, **kwargs):
        try:
            response = self.perform_dispatch(request, *args, **kwargs)
        except ActionError as e:
            response = self.render_error(e.message, 400)
        except AuthenticationError as e:
            response = self.render_error(e.message, 401)
        except DataConflict as e:
            response = self.render_error(e.message, 409)
        except FieldError as e:
            response = self.render_error(e.message, 400)
        except NotFound as e:
            response = self.render_error(e.message, 404)
        except ValidationError as e:
            response = self.render_error(e.message, 422)
        return response

    def perform_dispatch(self, request, *args, **kwargs):
        try:
            handler = self.get_handler(request, *args, **kwargs)
            self.check_permissions()
            self.set_bundle_from_request(request)
            response = handler(request, *args, **kwargs)
        except ObjectDoesNotExist as e:
            if self.model and not isinstance(e, self.model.DoesNotExist):
                raise e
            raise NotFound from e
        except RequestDataTooBig as e:
            self.request._body = self.request.read(None)  # prevent further raises
            max_upload_size = filesizeformat(settings.DATA_UPLOAD_MAX_MEMORY_SIZE)
            raise ValidationError(f"Max upload size is {max_upload_size}") from e
        return response

    def render_error(self, message, status):
        return self.render_to_response(dict(message=message), status)

    def check_permissions(self):
        try:
            for perm in self.permissions:
                perm()(self.request, **self.kwargs)
        except WorfError as e:
            if settings.WORF_DEBUG:  # pragma: no cover
                raise PermissionsError(
                    f"Permission check {perm.__module__}.{perm.__name__} raised {e.__class__.__name__}. "
                    f"You'd normally see a 4xx here but WORF_DEBUG=True."
                ) from e
            raise e

    def get_handler(self, request, *args, **kwargs):
        method = request.method.lower()
        handler = self.http_method_not_allowed
        if method in self.http_method_names:
            handler = getattr(self, method, self.http_method_not_allowed)
        return handler

    def get_instance(self):
        return self.instance if hasattr(self, "instance") else None

    def flatten_bundle(self, raw_bundle):
        # parse_qs gives us a dictionary where all values are lists
        return {
            key: value[0] if len(value) == 1 else value
            for key, value in raw_bundle.items()
        }

    def get_parts(self, request):
        if request.method == "POST":
            return request.POST, request.FILES

        return request.parse_file_upload(request.META, BytesIO(request.body))

    def set_bundle(self, raw_bundle, allow_negations=False):
        self.bundle = {}
        self.keymap = {}

        if not raw_bundle:
            return

        for key in raw_bundle.keys():
            field = (
                camel_to_snake(key[:-1]) + "!"
                if allow_negations and key.endswith("!")
                else camel_to_snake(key)
            )
            self.bundle[field] = raw_bundle[key]
            self.keymap[field] = key

    def set_bundle_from_request(self, request):
        if request.method == "GET":
            self.set_bundle_from_query_string(request)
            return

        self.set_bundle_from_request_body(request)

    def set_bundle_from_query_string(self, request):
        raw_bundle = self.flatten_bundle(parse_qs(request.META["QUERY_STRING"]))

        strings = dict(true=True, false=False, null=None)

        coerced_bundle = {
            key: strings.get(str(value).lower(), value)
            for key, value in raw_bundle.items()
        }

        self.set_bundle(coerced_bundle, allow_negations=True)

    def set_bundle_from_request_body(self, request):
        raw_bundle = {}

        if request.content_type == "multipart/form-data":
            post, files = self.get_parts(request)
            raw_bundle.update(self.flatten_bundle(post))
            raw_bundle.update(self.flatten_bundle(files))
        elif request.body:
            try:
                raw_bundle = json.loads(request.body)
            except json.decoder.JSONDecodeError:
                pass

        self.set_bundle(raw_bundle)



from worf.conf import settings


def test_settings():
    assert settings.WORF_API_NAME == "Test API"
    assert settings.WORF_API_ROOT == "/api/"
    assert settings.WORF_BROWSABLE_API is True
    assert settings.WORF_DEBUG is False



from worf.lookups import FindInstance
from worf.views.base import AbstractBaseAPI


class DeleteAPI(FindInstance, AbstractBaseAPI):
    def delete(self, *args, **kwargs):
        self.destroy(*args, **kwargs)
        return self.render_to_response("", 204)

    def destroy(self, *args, **kwargs):
        self.get_instance().delete()



from hashlib import md5
from uuid import uuid4

from django.contrib.auth.models import User
from django.db import models
from django.utils.timezone import now


class Profile(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid4)

    avatar = models.FileField(upload_to="avatars/", blank=True)
    email = models.EmailField(blank=True, max_length=320, null=True, unique=True)
    phone = models.CharField(max_length=32)

    boolean = models.BooleanField(blank=True, null=True)
    integer = models.IntegerField(blank=True, null=True)
    json = models.JSONField(blank=True, null=True)
    positive_integer = models.PositiveIntegerField(blank=True, null=True)
    slug = models.SlugField(blank=True, null=True)
    small_integer = models.SmallIntegerField(blank=True, null=True)

    user = models.OneToOneField(User, on_delete=models.CASCADE)
    role = models.ForeignKey("Role", on_delete=models.CASCADE)
    team = models.ForeignKey("Team", blank=True, null=True, on_delete=models.SET_NULL)
    skills = models.ManyToManyField("Skill", through="RatedSkill")
    tasks = models.ManyToManyField("Task")
    tags = models.ManyToManyField("Tag")

    recovery_email = models.EmailField(blank=True, max_length=320, null=True)
    recovery_phone = models.CharField(blank=True, max_length=32, null=True)
    resume = models.FileField(upload_to="resumes/", blank=True)

    last_active = models.DateField(blank=True, null=True)
    created_at = models.DateTimeField(blank=True, null=True)

    is_subscribed = models.BooleanField(blank=True, null=True)
    subscribed_at = models.DateTimeField(blank=True, null=True)
    subscribed_by = models.ForeignKey(
        User, blank=True, null=True, on_delete=models.SET_NULL
    )

    def get_avatar_url(self):
        return self.avatar.url if self.avatar else self.get_gravatar_url()

    def get_gravatar_hash(self):
        return md5(self.user.email.lower().encode()).hexdigest()

    def get_gravatar_url(self, default="identicon", size=512):
        return f"https://www.gravatar.com/avatar/{self.get_gravatar_hash()}?d={default}&s={size}"

    def subscribe(self, user=None):
        self.is_subscribed = True
        self.subscribed_at = now()
        self.subscribed_by = user
        self.save()

    def unsubscribe(self):
        self.is_subscribed = False
        self.subscribed_at = None
        self.subscribed_by = None
        self.save()


class Role(models.Model):
    name = models.CharField(max_length=100)


class Skill(models.Model):
    name = models.CharField(max_length=200)


class RatedSkill(models.Model):
    class Meta:
        ordering = ["skill__name"]
        unique_together = ["profile", "skill"]

    RATING_CHOICES = [
        (5, "Excellent"),
        (4, "Great"),
        (3, "Average"),
        (2, "Poor"),
        (1, "Bad"),
    ]

    profile = models.ForeignKey("Profile", on_delete=models.CASCADE)
    rating = models.IntegerField(choices=RATING_CHOICES)
    skill = models.ForeignKey("Skill", on_delete=models.CASCADE)

    def __str__(self):
        return f"{self.skill.name}: {self.get_rating_display()}"


class Tag(models.Model):
    name = models.CharField(max_length=200)


class Task(models.Model):
    custom_id = models.UUIDField(default=uuid4)
    name = models.CharField(max_length=200)

    class Api:
        lookup_field = "custom_id"


class Team(models.Model):
    name = models.CharField(max_length=100)


class UserMixin:
    @property
    def name(self):
        return f"{self.first_name} {self.last_name}".strip() or "Unknown User"


User.__bases__ += (UserMixin,)
